from .connect_without_apikey import KiteConnectBrowser
from .utils import get_totp
from .repeat_timer import RepeatTimer
import logging
import pickle
from datetime import datetime
import os
from typing import Optional
from kiteconnect import KiteConnect
from kiteconnect.exceptions import KiteException
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from urllib.parse import urlparse
from urllib.parse import parse_qs
from smart_open import open

logger = logging.getLogger(__name__)


class ZerodhaLoginException(Exception):
    """
    Exception raised while logging into Zerodha.
    """

    def __init__(self, config, error=None, message="Failed to login to Zerodha!"):
        super().__init__("{}\nProvided config: {}\n{}".format(message, config, "Error from client: {}".format(
            error) if error is not None else ""))


class KiteConnectionManager:
    """
    This class is responsible to login to Zerodha and return the KiteConnect handle through which we can act upon
    the particular account.

    @:param user_details dictionary containing login credentials - user_name, password, pin, google_authenticator_secret
        ,api_key, api_secret
    @:param refresh_connection: if true refresh the connection every refresh_interval_minutes
    @:param session_dir: directory where sessions would be saved to prevent re-login during restarts
    """
    __DATE_TODAY = datetime.now().strftime('%Y_%m_%d')

    def __init__(self, user_details: dict, refresh_connection: bool = False, refresh_interval_minutes: int = 10,
                 session_base_dir: str = os.getcwd()):
        if 'user_name' not in user_details or 'password' not in user_details:
            raise ZerodhaLoginException(config=user_details, message="Must have both user_name and password set")
        if 'pin' not in user_details and 'google_authenticator_secret' not in user_details:
            raise ZerodhaLoginException(config=user_details,
                                        message="One of pin or google_authenticator_secret must be provided!")
        self.config = user_details
        self.user_name = user_details.get('user_name', None)
        self.password = user_details.get('password', None)
        self.api_key = user_details.get('api_key', None)
        self.api_secret = user_details.get('api_secret', None)
        self.pin = user_details.get('pin', None)
        self.google_authenticator_secret = user_details.get('google_authenticator_secret', None)
        self.kite: KiteConnect = None
        self.use_api_key = self.api_key is not None and self.api_secret is not None
        self.session_dir = '{dir}/{date}'.format(dir=session_base_dir, date=KiteConnectionManager.__DATE_TODAY)
        if refresh_connection:
            self.background_thread = RepeatTimer(refresh_interval_minutes * 60, self.get_kite_connect)

    def __get_session_file_path(self):
        return '{dir}/{user}_{login_type}.session'.format(dir=self.session_dir,
                                                          user=self.user_name,
                                                          login_type='api' if self.is_using_api_key() else 'browser')

    def is_using_api_key(self) -> bool:
        return self.use_api_key

    def is_logged_in(self) -> bool:
        """
        Since, KiteConnect doesn't provide a clean way to check if the access token is valid. We will try to fetch the
        price of an instrument and see if a valid response is return. If we see a status code of 403 , it means that
        the session is expired. Check https://kite.trade/docs/connect/v3/exceptions/#common-http-error-codes to learn
        about error codes.

        @:return True if the KiteConnect is still valid
        """
        if not self.kite:
            if not self.__load_session():
                return False
        # If using API KEY, the access token stays valid for a day so no need to check for login
        if self.is_using_api_key():
            return True
        try:
            self.kite.ltp("NSE:INFY")
        except KiteException as ke:
            if ke.code == 403:
                return False
            raise ke
        return True

    def shutdown(self, invalidate_access_token=False):
        if self.background_thread:
            self.background_thread.cancel()
        # TODO: Need to be adapted for GCS
        # if invalidate_access_token and self.is_using_api_key():
        #     logger.info('Invalidating access_token for the user %s', self.user_name)
        #     self.kite.invalidate_access_token()
        #     logger.info('Clearing up session file for user %s at %s', self.user_name, self.__get_session_file_path())
        #     os.remove(self.__get_session_file_path())

    def __dump_session(self):
        with open(self.__get_session_file_path(), 'wb') as handle:
            pickle.dump(self.kite, handle, protocol=pickle.HIGHEST_PROTOCOL)
            logger.info('Dumped KiteConnect object for user %s at %s', self.user_name, self.__get_session_file_path())

    def __load_session(self) -> bool:
        file_path = self.__get_session_file_path()
        try:
            with open(file_path, 'rb') as handle:
                self.kite = pickle.load(handle)
                # if self.is_using_api_key():
                #     self.kite.renew_access_token(refresh_token=self.kite.access_token, api_secret=self.api_secret)
                #     self.__dump_session()
                return True
        except Exception as ioe:
            logger.info('Unable to load KiteConnect object for user %s from %s. Error: %s', self.user_name,
                        file_path, ioe)
            return False

    def get_kite_connect(self) -> KiteConnect:
        if not self.is_logged_in():
            self.kite = self.__login()
            self.__dump_session()
        return self.kite

    def __login(self) -> KiteConnect:
        """
        Login a user and fetch the corresponding KiteConnect handle
        """
        logger.info('Trying to login user_id %s', self.user_name)
        if self.is_using_api_key():
            return self.__login_with_apikey()
        return self.__login_without_apikey()

    def __login_without_apikey(self) -> KiteConnect:
        kite_connect_browser = KiteConnectBrowser(user_id=self.user_name, password=self.password,
                                                  google_authenticator_secret=self.google_authenticator_secret,
                                                  pin=self.pin)
        try:
            kite_connect_browser.login()
        except Exception as e:
            raise ZerodhaLoginException(self.config, e)
        logger.info("Successfully logged in %s without API Key", self.user_name)
        return kite_connect_browser

    def __login_with_apikey(self) -> KiteConnect:
        """
        Establishes connection to Kite by "always" generating a new session.
        """
        chrome_options = webdriver.ChromeOptions()
        chrome_options.add_argument('--no-sandbox')
        chrome_options.add_argument('--headless')
        chrome_options.add_argument('--disable-gpu')
        chrome_options.add_argument('--disable-dev-shm-usage')
        chrome_options.add_argument("--window-size=1920,1080")
        # from selenium.webdriver.chrome.service import Service
        # from webdriver_manager.chrome import ChromeDriverManager
        # driver = webdriver.Chrome(options=chrome_options, service=Service(ChromeDriverManager().install()))
        driver = webdriver.Chrome('/usr/local/bin/chromedriver', options=chrome_options)
        driver.implicitly_wait(5)
        try:
            import time
            kite = KiteConnect(api_key=self.api_key)
            driver.get(kite.login_url())

            WebDriverWait(driver=driver, timeout=10).until(
                EC.visibility_of_element_located((By.XPATH, "//div[@class='login-form']")))
            driver.find_element(By.XPATH, "//input[@type='text']").send_keys(self.user_name)
            driver.find_element(By.XPATH, "//input[@type='password']").send_keys(self.password)
            time.sleep(0.5)
            driver.find_element(By.XPATH, "//button[@type='submit']").click()

            time.sleep(0.5)
            WebDriverWait(driver=driver, timeout=10).until(
                EC.visibility_of_element_located((By.XPATH, "//div[@class='login-form']")))
            driver.find_element(By.XPATH, "//input[@type='text']").send_keys(
                get_totp(self.google_authenticator_secret) if self.pin is None else self.pin)

            time.sleep(0.5)
            request_token = KiteConnectionManager.__extract_request_token(driver.current_url)
            kite.generate_session(request_token, api_secret=self.api_secret)
            logger.info("Successfully logged in %s using API Key %s. Request Token: %s", self.user_name, self.api_key,
                        request_token)
            return kite
        except Exception as e:
            raise ZerodhaLoginException(self.config, e)
        finally:
            driver.close()
            driver.quit()

    @staticmethod
    def __extract_request_token(redirect_url: str) -> str:
        """
        Extracts request_token from the redirect_url
        :param redirect_url: The URL to which Kite API redirects to after successful login
        """
        try:
            parsed_url = urlparse(redirect_url)
            return parse_qs(parsed_url.query)['request_token'][0]
        except Exception as e:
            logger.error('Unable to extract request token from the redirect URL %s', redirect_url)



import unittest
from kite_connection_manager.connection_manager import KiteConnectionManager

class MyTestCase(unittest.TestCase):
    def test_something(self):
        creds = {
            "name": "Aneesha Kaushal",
            "user_name": "JFA773",
            "password": "buy@Farm1",
            "api_key": "97wr0exmsqicnu01",
            "api_secret": "ntos88g38uma3jkyczhg7ac6g8ar4pj5",
            "google_authenticator_secret": "7JCJA4VDGVCWUGDPRKVK3RGYVOB3F3EF",
            "quantity_multiplier": 0.5
        }
        connection_manager = KiteConnectionManager(user_details=creds, refresh_connection=False)
        connection_manager.get_kite_connect()
        self.assertEqual(True, False)


if __name__ == '__main__':
    unittest.main()



from .utils import get_totp
import logging
import json
import requests
from six.moves.urllib.parse import urljoin
from kiteconnect import KiteConnect, KiteTicker
import kiteconnect.exceptions as ex
from bs4 import BeautifulSoup

log = logging.getLogger(__name__)

base_url = "https://kite.zerodha.com"
login_url = "https://kite.zerodha.com/api/login"
twofa_url = "https://kite.zerodha.com/api/twofa"
instruments_url = "https://api.kite.trade/instruments"


class KiteConnectBrowser(KiteConnect):
    """
    Code borrowed from: https://github.com/jugaad-py/jugaad-trader/blob/master/jugaad_trader/zerodha.py
    """
    _default_root_uri = "https://kite.zerodha.com"

    def __init__(self, user_id=None, password=None, pin=None, google_authenticator_secret=None):

        self.user_id = user_id
        self.password = password
        self.pin = pin
        self.google_authenticator_secret = google_authenticator_secret

        super().__init__(api_key="")
        self.s = self.reqsession = requests.Session()
        headers = {
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.164 Safari/537.36",
            "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
            "accept-language": "en-GB,en-US;q=0.9,en;q=0.8",
        }
        self.reqsession.headers.update(headers)
        self.chunkjs = {}
        self.url_patch = '/oms'
        # self._routes["user.profile"] = "/user/profile/full"

    def set_access_token(self):
        self.login()

    def _user_agent(self):
        return "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.164 Safari/537.36"

    def login_step1(self):
        self.r = self.reqsession.get(base_url)
        self.r = self.reqsession.post(login_url, data={"user_id": self.user_id, "password": self.password})
        j = json.loads(self.r.text)
        return j

    def login_step2(self, j):
        data = {"user_id": self.user_id, "request_id": j['data']["request_id"],
                "twofa_value": get_totp(self.google_authenticator_secret) if self.pin is None else self.pin}
        self.r = self.s.post(twofa_url, data=data)
        j = json.loads(self.r.text)
        return j

    def login(self):
        j = self.login_step1()
        if j['status'] == 'error':
            raise Exception(j['message'])

        j = self.login_step2(j)
        if j['status'] == 'error':
            raise Exception(j['message'])
        self.enc_token = self.r.cookies['enctoken']
        return j

    def custom_headers(self):
        h = {}
        h['authorization'] = "enctoken {}".format(self.enc_token)
        h['referer'] = 'https://kite.zerodha.com/dashboard'
        h['x-kite-version'] = '2.9.2'
        h['sec-fetch-site'] = 'same-origin'
        h['sec-fetch-mode'] = 'cors'
        h['sec-fetch-dest'] = 'empty'
        h['x-kite-userid'] = self.user_id
        return h

    def _request(self, route, method, url_args=None, params=None,
                 is_json=False, query_params=None):
        if url_args:
            uri = self._routes[route].format(**url_args)
        else:
            uri = self._routes[route]

        url = urljoin(self.root, self.url_patch + uri)

        # prepare url query params
        if method in ["GET", "DELETE"]:
            query_params = params

        # Custom headers
        headers = self.custom_headers()

        if self.debug:
            log.debug("Request: {method} {url} {params} {headers}".format(method=method, url=url, params=params,
                                                                          headers=headers))

        try:
            r = self.reqsession.request(method,
                                        url,
                                        json=params if (method in ["POST", "PUT"] and is_json) else None,
                                        data=params if (method in ["POST", "PUT"] and not is_json) else None,
                                        params=query_params,
                                        headers=headers,
                                        verify=not self.disable_ssl,
                                        allow_redirects=True,
                                        timeout=self.timeout,
                                        proxies=self.proxies)
            self.r = r
        except Exception as e:
            raise e

        if self.debug:
            log.debug("Response: {code} {content}".format(code=r.status_code, content=r.content))

        # Validate the content type.
        if "json" in r.headers["content-type"]:
            try:
                data = json.loads(r.content.decode("utf8"))
            except ValueError:
                raise ex.DataException("Couldn't parse the JSON response received from the server: {content}".format(
                    content=r.content))

            # api error
            if data.get("status") == "error" or data.get("error_type"):
                # Call session hook if its registered and TokenException is raised
                if self.session_expiry_hook and r.status_code == 403 and data["error_type"] == "TokenException":
                    self.session_expiry_hook()

                # native Kite errors
                exp = getattr(ex, data.get("error_type"), ex.GeneralException)
                raise exp(data["message"], code=r.status_code)

            return data["data"]
        elif "csv" in r.headers["content-type"]:
            return r.content
        else:
            raise ex.DataException("Unknown Content-Type ({content_type}) with response: ({content})".format(
                content_type=r.headers["content-type"],
                content=r.content))

    def get_chunk_js(self):
        self.r = self.reqsession.get(urljoin(base_url, '/dashboard'))
        html = self.r.text
        bs = BeautifulSoup(html)
        for tag in bs.find_all("link"):
            src = tag.attrs.get("href", "")
            if "chunk" in src:
                break
        url = urljoin(base_url, tag.attrs.get("href"))
        self.r = self.reqsession.get(url)
        return self.r.text

    def chunk_to_json(self, js):
        start = js.find('{"months"')
        end = js.find("\')}}])")
        jtxt = js[start:end].replace('\\', '')
        self.chunkjs = json.loads(jtxt)
        return self.chunkjs

    def instruments(self, exchange=None):
        if exchange:
            self.r = self.reqsession.get(instruments_url + "/{}".format(exchange))
        else:
            self.r = self.reqsession.get(instruments_url)
        return self._parse_instruments(self.r.text)

    def close(self):
        self.reqsession.close()

    def ticker(self, api_key='kitefront', enctoken=None, userid=None):
        if enctoken is not None:
            self.enctoken = self.r.cookies['enc_token']
        if userid is not None:
            self.user_id = self.user_id
        if self.user_id is None:
            raise Exception(
                "\nCould not find the session, Please start a session using \n\n$ jtrader zerodha startsession")
        return KiteTicker(api_key=api_key, access_token=self.enc_token + '&user_id=' + self.user_id,
                          root='wss://ws.zerodha.com')


class Console(KiteConnectBrowser):
    """
        Experimental support for Zerodha backoffice platform Coin
    """
    _default_root_uri = "https://console.zerodha.com"
    api_key = "console"  # API key for Coin
    _routes = {
        "login": "/kite/login",
        "status": "/api/user/status",
        "user_profile": "/api/user_profile/status",
        "dashboard": "/api/dashboard",
        "account_values": "/api/dashboard/account_values",
        "positions": "/api/reports/positions",
        "exposure": "/api/reports/positions/exposure",
        "portfolio": "/api/reports/holdings/portfolio",
        "tradebook": "/api/reports/tradebook",
        "pnl": "/api/reports/pnl",
        "pnl_summary": "/api/reports/pnl/summary",
        "tax_pnl": "/api/reports/taxpnl",
        "fund_balance": "https://console.zerodha.com/api/funds/balance",
        "ledger": "/api/ledger",
        "interest_statement": "/api/funds/interest_statement",
        "mandate": "/api/mandate"
    }

    def __init__(self, z):
        """
            args:
                z - Instance of Zerodha class
        """

        super().__init__(z.user_id, z.password, z.pin, z.google_authenticator_secret)
        self._root = self._default_root_uri

        self.url_patch = ""
        self.reqsession = z.reqsession
        headers = {
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.164 Safari/537.36",
            "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
            "accept-language": "en-GB,en-US;q=0.9,en;q=0.8",
        }
        self.reqsession.headers.update(headers)
        self.console_session = ""
        self.public_token = ""
        self.register_functions()

    def custom_headers(self):
        h = {}
        h['referer'] = 'https://console.zerodha.com/'
        h['x-kite-version'] = '3'
        h['sec-fetch-site'] = 'same-origin'
        h['sec-fetch-mode'] = 'cors'
        h['sec-fetch-dest'] = 'empty'
        h[
            'user-agent'] = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.164 Safari/537.36"

        return h

    def login(self):
        """
            Get request to /login redirects to kite, if the Kite session is
            already active, You are redirected back to Console and logged in
            automatically
        """
        url = self._root + self._routes["login"]
        self.r = self.reqsession.get(url)
        if self.r.url == 'https://console.zerodha.com/dashboard':
            cookies = self.reqsession.cookies.get_dict('console.zerodha.com')
            self.console_session = cookies['session']
            return True
        else:
            raise Exception("Login failed or Kite session expired")

    def factory_functions(self, route, docstring=""):
        """
            All APIs used by console end up sendinga a get request with some
            data, hence creating a factory function which will generate a
            function to send GET requests
        """

        def generic_function(**kwargs):
            return self._get(route, params=kwargs)

        generic_function.__doc__ = docstring
        return generic_function

    def register_functions(self):
        self.dashboard = self.factory_functions("dashboard")
        self.account_values = self.factory_functions("account_values")
        self.positions = self.factory_functions("positions")
        self.exposure = self.factory_functions("exposure")
        self.portfolio = self.factory_functions("portfolio")
        self.tradebook = self.factory_functions("tradebook")
        self.pnl = self.factory_functions("pnl")
        self.pnl_summary = self.factory_functions("pnl_summary")
        self.tax_pnl = self.factory_functions("tax_pnl")
        self.fund_balance = self.factory_functions("fund_balance")
        self.ledger = self.factory_functions("ledger")
        self.interest_statement = self.factory_functions("interest_statement")
        self.mandate = self.factory_functions("mandate")


if __name__ == "__main__":
    pass



from threading import Timer


class RepeatTimer(Timer):
    """
    Repeat the timer until cancelled
    """

    def run(self):
        while not self.finished.wait(self.interval):
            self.function(*self.args, **self.kwargs)



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='kite_connection_manager',
    packages=find_packages(),
    version='0.4.6',
    description='Manage Kite Connections',
    author='Shashwat Rastogi',
    author_email='shashwat1991@gmail.com',
    license='MIT',
    install_requires=['pyotp~=2.6.0', 'kiteconnect~=4.1.0', 'selenium~=4.3.0', 'websockets', 'requests',
                      'beautifulsoup4', 'six', 'webdriver-manager', 'smart_open[gcs]'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest==4.4.1'],
    test_suite='tests',
    long_description=long_description,
    url="https://github.com/shashwat7/KiteConnectionManager",
)
